# import matplotlib.pyplot as plt
from typing import Tuple, Union
import numpy as np


# TODO: Implement the plot_magfield function
def plot_magfield(
        magnetic_field: Tuple[Union[np.ndarray, list]],
        labels: list[str],
        title: str = "Magnetic Field",
        xlabel: str = "Time",
        ylabel: str = "Magnetic Field ",
):
    pass


# TODO: Implement the plot_attitude function
def plot_three_axes(
        time: Tuple[Union[np.ndarray, list]],
        data: Tuple[Union[np.ndarray, list]],
        labels: list[str],
        title: str = "Three Axes",
        xlabel: str = "Time",
        ylabel: str = "Three Axes",
):
    pass
